from __future__ import annotations

import asyncio

import pytest

import luxos.asyncops as aapi
from luxos import exceptions, misc

## NOTE ##
# This tests spawn an underlying server, it might be better not run
# unattended. Some also require a miner, we might not have it handy.
pytestmark = pytest.mark.manual


def test_parameters_to_list():
    assert aapi.parameters_to_list(None) == []
    assert aapi.parameters_to_list(0) == ["0"]
    assert aapi.parameters_to_list([0]) == ["0"]
    assert aapi.parameters_to_list(["0"]) == ["0"]
    assert aapi.parameters_to_list("hello") == ["hello"]
    assert aapi.parameters_to_list(["hello", "world"]) == ["hello", "world"]
    assert aapi.parameters_to_list(["hello", 1]) == ["hello", "1"]
    assert aapi.parameters_to_list({"hello": 1}) == ["hello=1"]
    assert aapi.parameters_to_list({"hello": True}) == ["hello=true"]


def test_validate_message():
    pytest.raises(
        exceptions.MinerCommandMalformedMessageError, aapi.validate_message, "a", 0, {}
    )
    host, port = "a", 0

    assert aapi.validate_message(host, port, {"STATUS": 1, "id": 2})

    pytest.raises(
        exceptions.MinerCommandMalformedMessageError,
        aapi.validate_message,
        host,
        port,
        {
            "STATUS": 1,
            "id": 2,
        },
        "wooow",
    )

    assert aapi.validate_message(
        host, port, {"STATUS": 1, "id": 2, "wooow": [1, 2]}, "wooow", 2, 2
    )

    with pytest.raises(exceptions.MinerCommandMalformedMessageError) as excinfo:
        aapi.validate_message(
            host, port, {"STATUS": 1, "id": 2, "wooow": [1, 2]}, "wooow", minfields=9
        )
    assert excinfo.value.args[2] == "found 2 fields for wooow invalid: 2 <= 9"

    with pytest.raises(exceptions.MinerCommandMalformedMessageError) as excinfo:
        aapi.validate_message(
            host, port, {"STATUS": 1, "id": 2, "wooow": [1, 2]}, "wooow", maxfields=1
        )
    assert excinfo.value.args[2] == "found 2 fields for wooow invalid: 2 >= 1"

    with pytest.raises(exceptions.MinerCommandMalformedMessageError) as excinfo:
        aapi.validate_message(
            host,
            port,
            {"STATUS": 1, "id": 2, "wooow": [1, 2]},
            "wooow",
            minfields=9,
            maxfields=10,
        )
    assert excinfo.value.args[2] == "found 2 fields for wooow invalid: 2 <= 9"

    assert aapi.validate_message(
        host,
        port,
        {"STATUS": 1, "id": 2, "wooow": [1, 2]},
        "wooow",
        minfields=2,
        maxfields=10,
    )


@pytest.mark.asyncio
async def test_private_roundtrip_one_listener(echopool):
    """checks roundrtip sends and receive a message (1-listener)"""
    echopool.start(1, mode="echo+")
    host, port = echopool.addresses[0]
    ret = await aapi._roundtrip(host, port, "hello", None)
    assert ret == f"received by ('{host}', {port}): hello"


@pytest.mark.asyncio
async def test_private_roundtrip_many_listeners(echopool):
    """checks the roundrip can connect en-masse to many lsiteners"""
    echopool.start(100, mode="echo+")

    blocks = {}
    for index, group in enumerate(misc.batched(echopool.addresses, 10)):
        tasks = []
        for host, port in group:
            tasks.append(aapi._roundtrip(host, port, "hello olleh", None))
        blocks[index] = await asyncio.gather(*tasks, return_exceptions=True)

    assert len(blocks) == 10
    assert sum(len(x) for x in blocks.values()) == 100
    allitems = [item for b in blocks.values() for item in b]
    assert f"received by {echopool.addresses[3]}: hello olleh" in allitems


@pytest.mark.asyncio
async def test_miner_logon_logoff_cycle(miner_host_port):
    host, port = miner_host_port

    sid = None
    try:
        sid = await aapi.logon(host, port)
    except exceptions.MinerCommandSessionAlreadyActive as e:
        raise RuntimeError("a session is already active on {host}:{port}") from e
    finally:
        if sid:
            await aapi.logoff(host, port, sid)


@pytest.mark.asyncio
async def test_miner_double_logon_cycle(miner_host_port):
    host, port = miner_host_port

    sid = await aapi.logon(host, port)
    try:
        with pytest.raises(exceptions.MinerCommandSessionAlreadyActive) as excinfo:
            await aapi.logon(host, port)
        assert excinfo.value.args[0] == f"session active for {host}:{port}"
    except Exception:
        pass
    finally:
        if sid:
            await aapi.logoff(host, port, sid)


@pytest.mark.asyncio
async def test_miner_version(miner_host_port):
    host, port = miner_host_port

    res = await aapi.rexec(host, port, "version")
    assert "VERSION" in res
    assert len(res["VERSION"]) == 1
    assert "API" in res["VERSION"][0]


# @pytest.mark.skip("a likely bug in luxminer way to set/delete profiles")
@pytest.mark.asyncio
async def test_miner_profile_sets(miner_host_port):
    from random import choices
    from string import ascii_lowercase

    # random profile name
    profile = f"test-{''.join(choices(ascii_lowercase, k=5))}"

    # get the initial profile list
    host, port = miner_host_port
    profiles = (await aapi.rexec(host, port, "profiles"))["PROFILES"]
    assert profiles and profile not in {p["Profile Name"] for p in profiles}

    # create a new profile
    # TODO there's a bug
    #  when ATM is running you shouldn't be able to create/delete
    #  profiles, the following test should fail at the first try
    # -> to pass this test, you need to disable ATM

    params = f"{profile},{profiles[0]['Frequency']},{profiles[0]['Voltage']}"
    await aapi.rexec(host, port, "profilenew", params)

    try:
        profiles1 = (await aapi.rexec(host, port, "profiles"))["PROFILES"]
        assert profile in {p["Profile Name"] for p in profiles1}
    finally:
        async with aapi.with_atm(host, port, False):
            await aapi.rexec(host, port, "profilerem", profile)

    # verify we restored the same profiles
    profiles2 = (await aapi.rexec(host, port, "profiles"))["PROFILES"]
    expected = {p["Profile Name"] for p in profiles}
    found = {p["Profile Name"] for p in profiles2}
    assert found == expected


@pytest.mark.asyncio
async def test_roundtrip_timeout(miner_host_port):
    # host, port = miner
    """checks roundrtip sends and receive a message (1-listener)"""

    # miner doesn't exist
    host, port = "127.0.0.99", 12345
    try:
        await aapi.rexec(host, port, "hello", timeout=0.5)
        assert False, "didn't raise"
    except aapi.exceptions.MinerCommandTimeoutError as exc:
        exception = exc
    assert isinstance(exception, asyncio.TimeoutError)
    assert isinstance(exception, aapi.exceptions.MinerConnectionError)
    assert isinstance(exception, aapi.exceptions.LuxosBaseException)
    text = f"<{host}:{port}>: MinerCommandTimeoutError, TimeoutError()"
    assert str(exception)[: len(text)] == text

    # not miner on a wrong port
    host, port = miner_host_port
    port = 22
    exception = None
    try:
        await aapi.rexec(host, port, "hello", timeout=0.5)
        assert False, "didn't raise"
    except aapi.exceptions.MinerCommandTimeoutError as exc:
        exception = exc
    assert isinstance(exception, asyncio.TimeoutError)
    assert isinstance(exception, aapi.exceptions.MinerConnectionError)
    assert isinstance(exception, aapi.exceptions.LuxosBaseException)
    texts = [
        f"<{host}:{port}>: MinerCommandTimeoutError, ConnectionRe",
        f"<{host}:{port}>: MinerCommandTimeoutError, TimeoutError",
    ]
    assert str(exception)[: len(texts[0])] in texts


@pytest.mark.asyncio
async def test_bridge_execute_command(miner_host_port):
    from luxos.utils import execute_command, rexec

    # get the initial profile list
    host, port = miner_host_port

    out = execute_command(host, port, 3, "profiles", parameters=[], verbose=True)
    out1 = await rexec(host, port, cmd="profiles")
    assert out["PROFILES"] == out1["PROFILES"]

    port += 1
    pytest.raises(
        ConnectionRefusedError,
        execute_command,
        host,
        port,
        3,
        "profiles",
        parameters=[],
        verbose=True,
    )



"""various argparse `type` attributes"""

from __future__ import annotations

import argparse
import contextlib
import datetime
import logging
from pathlib import Path
from typing import Any, Sequence

from .shared import LuxosParserBase


def type_range(txt: str) -> Sequence[tuple[str, int | None]]:
    """type conversion for ranges

    This will enforce conversion between a string and a ranged object.

    Eg.
        parser.add_argument("--range", type=type_range)

        The --range argument will be:
            127.0.0.1  # single ip address
            127.0.0.1:1234  # single ip address with port
            127.0.0.1-127.0.0.3 # a list of (ip, port) tuple between *.1 and.3

        Alternatively you can pass a @filename to read data from a csv file
    """
    from luxos.ips import iter_ip_ranges, load_ips_from_csv, load_ips_from_yaml

    path = None
    if txt.startswith("@") and not (path := Path(txt[1:])).exists():
        raise argparse.ArgumentTypeError(f"file not found {path}")

    if path:
        with contextlib.suppress(RuntimeError):
            return load_ips_from_yaml(path)

    try:
        if path:
            return load_ips_from_csv(path)
        return list(iter_ip_ranges(txt))
    except RuntimeError as exc:
        raise argparse.ArgumentTypeError(f"conversion failed '{txt}': {exc.args[0]}")
    except Exception as exc:
        raise argparse.ArgumentTypeError(
            f"unknown conversion failed for '{txt}'"
        ) from exc


def type_hhmm(txt: str):
    """type conversion for ranges

    This will enforce conversion between a string and datetime.time object.

    Eg.
        parser.add_argument("--time", type=type_hhmm)

        The --time format is HH:MM
    """
    if not txt:
        return
    with contextlib.suppress(ValueError, TypeError):
        hh, _, mm = txt.partition(":")
        hh1 = int(hh)
        mm1 = int(mm)
        return datetime.time(hh1, mm1)
    raise argparse.ArgumentTypeError(f"failed conversion into HH:MM for '{txt}'")


def add_arguments_rexec(parser: LuxosParserBase):
    """adds the rexec timing for timeout/retries/delays

    Ex.

    def add_arguments(parser):
        cli.flags.add_arguments_rexec(parser)

    def process_args(args):
        asyncops.TIMEOUT = args.timeout
        asyncops.RETRIES = args.retries
        asyncops.RETRIES_DELAY = args.retries_delay
        return args
    """
    from ..asyncops import RETRIES, RETRIES_DELAY, TIMEOUT

    group = parser.add_argument_group(
        "Remote execution", "rexec remote execution limits/timeouts"
    )
    group.add_argument(
        "--timeout", type=float, default=TIMEOUT, help="Timeout for each command"
    )
    group.add_argument(
        "--retries",
        type=int,
        default=RETRIES,
        help="Maximum number of retries for each command",
    )
    group.add_argument(
        "--retries-delay",
        type=float,
        default=RETRIES_DELAY,
        help="Delay in s between retries",
    )

    def callback(args: argparse.Namespace):
        from .. import asyncops

        asyncops.TIMEOUT = args.timeout
        asyncops.RETRIES = args.retries
        asyncops.RETRIES_DELAY = args.retries_delay

    parser.callbacks.append(callback)


def add_arguments_config(parser: LuxosParserBase):
    # find the CONFIGPATH, from the module then from luxos.cli.v*
    default = None
    for module in reversed(parser.modules):
        if not hasattr(module, "CONFIGPATH"):
            continue
        default = Path.cwd() / Path(module.CONFIGPATH).expanduser()
        break
    if default and default.is_relative_to(Path.cwd()):
        default = default.relative_to(Path.cwd())

    parser.add_argument(
        "-c",
        "--config",
        default=default,
        type=Path,
        help="path to a config file",
    )

    def callback(args: argparse.Namespace):
        args.config = args.config.absolute() if args.config else 1

    parser.callbacks.append(callback)


def add_arguments_logging(parser: LuxosParserBase):
    group = parser.add_argument_group("Logging", "Logging related options")
    group.add_argument("-v", "--verbose", action="count", help="report verbose logging")
    group.add_argument("-q", "--quiet", action="count", help="report quiet logging")

    def setup_logging(config: dict[str, Any], count: int) -> None:
        levelmap = [
            logging.WARNING,
            logging.INFO,
            logging.DEBUG,
        ]
        n = len(levelmap)

        # awlays start from info level
        level = logging.INFO

        # we can set the default start log level in LOGGING_CONFIG
        if config.get("level", None) is not None:
            level = config["level"]

        # we control if we go verbose or quite here
        index = levelmap.index(level) + count
        config["level"] = levelmap[max(min(index, n - 1), 0)]
        logging.basicConfig(**config)

    def callback(args: argparse.Namespace):
        config = {}
        for module in reversed(args.modules):
            if not hasattr(module, "LOGGING_CONFIG"):
                continue
            config.update(module.LOGGING_CONFIG)

        count = (args.verbose or 0) - (args.quiet or 0)
        setup_logging(config, count)
        delattr(args, "verbose")
        delattr(args, "quiet")

    parser.callbacks.append(callback)


def add_arguments_database(parser: LuxosParserBase):
    """
    takes a string on a command line and retunr a sa engine.

    This is mean to be used as type in add_arguments

    .. parse_args
    args.engine = firmware.cli.process_engine(args.engine)

    The text format can be:
    - a string that will resolve into a Path
    - any string for sqlalchemy.engine.url.make_url

    In the first case a sqlite engine pointing to text Path will be returned,
    in the second a regular engine.

    Example:
    -------
        >>> process_engine("foobar.db")
        >>> process_engine("postgresql+psycopg2://<user>:<password>@<host>/<db>"

    """

    group = parser.add_argument_group("database", "Database related options")
    group.add_argument("--db", dest="engine", help="sqlalchemy uri or filename")
    group.add_argument(
        "--db-create", action="store_true", help="create the db structures"
    )

    def callback(args: argparse.Namespace):
        from sqlalchemy import create_engine
        from sqlalchemy.engine.url import make_url
        from sqlalchemy.exc import ArgumentError

        if not args.engine or not args.engine.strip():
            return None

        engine = None
        with contextlib.suppress(ArgumentError):
            url = make_url(args.engine)
            engine = create_engine(url)
        args.engine = engine or create_engine(f"sqlite:///{args.engine}", echo=False)

    parser.callbacks.append(callback)



"""ipaddress manipulation"""

from __future__ import annotations

import ipaddress
import re
from pathlib import Path
from typing import Generator

from .exceptions import LuxosBaseException


class DataParsingError(LuxosBaseException):
    pass


def splitip(txt: str) -> tuple[str, int | None]:
    expr = re.compile(r"(?P<ip>\d{1,3}([.]\d{1,3}){3})(:(?P<port>\d+))?")
    if not (match := expr.search(txt)):
        raise RuntimeError(f"invalid ip:port address {txt}")
    return match["ip"], int(match["port"]) if match["port"] is not None else None


def _parse_expr(txt: str) -> None | tuple[str, str, int | None]:
    tokens = {
        "ip": re.compile(r"(?P<ip>\d{1,3}([.]\d{1,3}){3})"),
        "sep": re.compile(":"),
        "div": re.compile("-"),
        "port": re.compile(r"(?P<port>\d+)"),
    }

    txt2 = txt.replace(" ", "")

    items = []
    while txt2.strip():
        for k, e in tokens.items():
            if match := e.match(txt2):
                i, j = match.span()
                items.append((k, txt2[i:j]))
                txt2 = txt2[j:]
                break
        else:
            raise RuntimeError(f"cannot parse text '{txt}'")

    if len(items) == 0:
        return None

    def matcher(syntax):
        def match(left, right):
            if len(left) != len(right):
                return False
            for a, b in zip(left, right):
                if isinstance(b, str):
                    if a != b:
                        return False
                elif a not in b:
                    return False
            return True

        start = end = port = None
        if match(syntax, ["ip"]):
            start = items[0][1]
        elif match(syntax, ["ip", "sep", "port"]):
            start = items[0][1]
            port = int(items[2][1])
        elif match(syntax, ["ip", {"sep", "div"}, "ip"]):
            start = items[0][1]
            end = items[2][1]
        elif match(syntax, ["ip", "sep", "port", {"div", "sep"}, "ip"]):
            start = items[0][1]
            end = items[4][1]
            port = int(items[2][1])
        elif match(syntax, ["ip", {"div", "sep"}, "ip", "sep", "port"]):
            start = items[0][1]
            end = items[2][1]
            port = int(items[4][1])
        elif match(syntax, ["ip", "sep", "port", {"div", "sep"}, "ip", "sep", "port"]):
            start = items[0][1]
            end = items[4][1]
            port = int(items[2][1])
            port1 = int(items[6][1])
            if port != port1:
                raise RuntimeError(f"ports mismatch {port} != {port1}")
        return start, end, port

    syntax = [item[0] for item in items]
    return matcher(syntax)


def iter_ip_ranges(
    txt: str, port: int | None = None, rsep: str = "-", gsep: str = ","
) -> Generator[tuple[str, int | None], None, None]:
    """iterate over ip ranges.

    The txt string cav have one of these formats:

    1. a single ip such as '127.0.0.1' or '127.0.0.1:8080'
    2. an (inclusive) range using two ips separated by `-`
         as '127.0.0.1 - 127.0.0.3'
    3. a combination of the above `,` separated as
         '127.0.0.1 , 192.168.0.1-192.168.0.10'

    Example:
    ```python
    for ip in iter_ip_ranges("127.0.0.1 , 127.0.0.3-127.0.0.15"):
        print(ip)

    (127.0.0.1, None),
    (127.0.0.2, None),
    (127.0.0.3, None),
    ...
    127.0.0.15
    ```
    """
    for segment in txt.replace(" ", "").split(gsep):
        if not (found := _parse_expr(segment)):
            continue
        start, end, theport = found
        if start is None and end is None:
            raise RuntimeError(f"cannot parse '{segment}'")
        if end is None:
            yield (start, theport or port)
            return

        cur = ipaddress.IPv4Address(start)
        last = ipaddress.IPv4Address(end)
        while cur <= last:
            yield (str(cur), theport or port)
            cur += 1


def ip_ranges(
    txt: str, rsep: str = "-", gsep: str = ":"
) -> list[tuple[str, int | None]]:
    """return a list of ips given a text expression.

    Eg.
        >>> for ip in ip_ranges("127.0.0.1"):
        ...     print(ip)
        127.0.0.1

        >>> for ip in ip_ranges("127.0.0.1-127.0.0.3"):
        ...     print(ip)
        127.0.0.1
        127.0.0.2
        127.0.0.3

    NOTE: use the `:` (gsep) to separate ips groups, and `-` (rsep) to define a range.
    """
    return list(iter_ip_ranges(txt, rsep=rsep, gsep=gsep))


def load_ips_from_csv(path: Path | str, port: int = 4028) -> list[tuple[str, int]]:
    """loads ip addresses from a csv file

    Example:
    ```python

    foobar.csv contains ranges as parsed by iter_ip_ranges
    127.0.0.1 # a single address
    127.0.0.2-127.0.0.10


    for ip in load_ips_from_csv("foobar.csv"):
        print(ip)

    (127.0.0.1, 4028)
    (127.0.0.2, 4028)
    (127.0.0.3, 4028)
    ...
    (127.0.0.10, 4028)
    ```

    """
    result = []
    for line in Path(path).read_text().split("\n"):
        line = line.partition("#")[0]
        if not line.strip():
            continue
        for host, port2 in iter_ip_ranges(line):
            result.append((host, port2 or port))
    return result


def load_ips_from_yaml(path: Path | str, port: int = 4028) -> list[tuple[str, int]]:
    """Loads ip addresses from a yaml file"""
    from yaml import safe_load

    txt = Path(path).read_text()
    try:
        data = safe_load(txt)
    except Exception as exc:
        raise DataParsingError(f"cannot parse yaml file {path}") from exc

    if "miners" in data and "addresses" in data["miners"]:
        miners = data["miners"]
        default_port = miners.get("luxos_port", None)
        result = []
        for address in data["miners"]["addresses"]:
            for host, thisport in iter_ip_ranges(address):
                result.append((host, thisport or default_port or port))
        return result

    raise DataParsingError(f"cannot find miners definitions in {path}")



"""user facing functions for normal use"""

from __future__ import annotations

import asyncio
import functools
import traceback
from typing import Any, Callable

import luxos.misc
from luxos.asyncops import rexec  # noqa: F401

# we bring here functions from other modules
from luxos.exceptions import (  # noqa: F401
    LuxosLaunchError,
    LuxosLaunchTimeoutError,
    MinerConnectionError,
)
from luxos.ips import ip_ranges, load_ips_from_csv  # noqa: F401
from luxos.syncops import execute_command  # noqa: F401


async def launch(
    addresses: list[tuple[str, int]], call: Callable[[str, int], Any], *args, **kwargs
) -> Any:
    """launch an async function on a list of (host, port) miners

    Special kwargs:
        - batch execute operation in group of batch tasks (rate limiting)
        - naked do not wrap the call, so is up to you catching exceptions (default None)

    Eg.
        async printme(host, port, value):
            print(await rexec(host, port, "version"))
        asyncio.run(launch([("127.0.0.1", 4028)], printme, value=11, batch=10))
    """

    # a naked options, wraps the 'call' and re-raise exceptions as LuxosLaunchError
    naked = kwargs.pop("naked") if "naked" in kwargs else None

    def wraps(fn):
        @functools.wraps(fn)
        async def _fn(host: str, port: int):
            try:
                return await fn(host, port)
            except asyncio.TimeoutError as exc:
                tback = "".join(traceback.format_exc())
                raise LuxosLaunchTimeoutError(tback, host, port) from exc
            except Exception as exc:
                tback = "".join(traceback.format_exc())
                raise LuxosLaunchError(tback, host, port) from exc

        return _fn

    n = int(kwargs.pop("batch") or 0) if "batch" in kwargs else None
    if n and n < 0:
        raise RuntimeError(
            f"cannot pass the 'batch' keyword argument with a value < 0: batch={n}"
        )
    if not naked:
        call = wraps(call)

    if n:
        result = []
        for subaddresses in luxos.misc.batched(addresses, n):
            tasks = [call(*address, *args, **kwargs) for address in subaddresses]
            for task in await asyncio.gather(*tasks, return_exceptions=True):
                result.append(task)
        return result
    else:
        tasks = [call(*address, *args, **kwargs) for address in addresses]
        return await asyncio.gather(*tasks, return_exceptions=True)



from __future__ import annotations

import argparse
import sys
import types
from typing import Callable

# The return type of add_arguments

if sys.version_info >= (3, 10):
    ArgsCallback = Callable[[argparse.Namespace], None | argparse.Namespace]
else:
    ArgsCallback = Callable


# The luxor base parser
class LuxosParserBase(argparse.ArgumentParser):
    def __init__(self, modules: list[types.ModuleType], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.modules = modules
        self.callbacks: list[ArgsCallback | None] = []



from __future__ import annotations

import contextlib
import json
import logging
import socket
from typing import Any

from luxos.api import logon_required

from .asyncops import TIMEOUT, parameters_to_list, validate_message
from .exceptions import MinerCommandSessionAlreadyActive, MinerConnectionError

log = logging.getLogger(__name__)


# internal_send_cgminer_command sends a command to the
# cgminer API server and returns the response.
def internal_send_cgminer_command(
    host: str, port: int, command: str, timeout: float | None
) -> dict[str, Any]:
    timeout_sec = TIMEOUT if timeout is None else timeout
    # Create a socket connection to the server
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        try:
            # set timeout
            if timeout_sec is not None:
                sock.settimeout(timeout)

            # Connect to the server
            sock.connect((host, port))

            # Send the command to the server
            sock.sendall(command.encode())

            # Receive the response from the server
            response = bytearray()
            while True:
                # Read one byte at a time so we can wait for the null terminator.
                # this is to avoid waiting for the timeout as we don't know how long
                # the response will be and socket.recv() will block until reading
                # the specified number of bytes.
                try:
                    data = sock.recv(1)
                except socket.timeout:
                    # Timeout occurred, check if we have any data so far
                    if len(response) == 0:
                        raise ValueError("timeout waiting for data")
                    else:
                        break
                if not data:
                    break
                null_index = data.find(b"\x00")
                if null_index >= 0:
                    response += data[:null_index]
                    break
                response += data

            # Parse the response JSON
            r = json.loads(response.decode())
            log.debug(r)
            return r

        except socket.error as e:
            raise e


# send_cgminer_command sends a command to the cgminer API server and
# returns the response.
def send_cgminer_command(
    host: str, port: int, cmd: str, param: str, timeout: float | None = None
) -> dict[str, Any]:
    timeout = TIMEOUT if timeout is None else timeout
    req = str(f'{{"command": "{cmd}", "parameter": "{param}"}}\n')
    log.debug(f"Executing command: {cmd} with params: {param} to host: {host}")
    return internal_send_cgminer_command(host, port, req, timeout)


# send_cgminer_simple_command sends a command with no params
# to the miner and returns the response.
def send_cgminer_simple_command(
    host: str, port: int, cmd: str, timeout: float | None = None
) -> dict[str, Any]:
    timeout = TIMEOUT if timeout is None else timeout
    req = str(f'{{"command": "{cmd}"}}\n')
    log.debug(f"Executing command: {cmd} to host: {host}")
    return internal_send_cgminer_command(host, port, req, timeout)


# check_res_structure checks that the response has the expected structure.
def check_res_structure(
    res: dict[str, Any], structure: str, min: int, max: int
) -> dict[str, Any]:
    # Check the structure of the response.
    if structure not in res or "STATUS" not in res or "id" not in res:
        raise ValueError("error: invalid response structure")

    # Should we check min and max?
    if min >= 0 and max >= 0:
        # Check the number of structure elements.
        if not (min <= len(res[structure]) <= max):
            raise ValueError(
                f"error: unexpected number of {structure} in response; "
                f"min: {min}, max: {max}, actual: {len(res[structure])}"
            )

    # Should we check only min?
    if min >= 0:
        # Check the minimum number of structure elements.
        if len(res[structure]) < min:
            raise ValueError(
                f"error: unexpected number of {structure} in response; "
                f"min: {min}, max: {max}, actual: {len(res[structure])}"
            )

    # Should we check only max?
    if max >= 0:
        # Check the maximum number of structure elements.
        if len(res[structure]) < min:
            raise ValueError(
                f"error: unexpected number of {structure} in response; "
                f"min: {min}, max: {max}, actual: {len(res[structure])}"
            )

    return res


# get_str_field tries to get the field as a string and returns it.
def get_str_field(struct: dict[str, Any], name: str) -> str:
    try:
        s = str(struct[name])
    except Exception as e:
        raise ValueError(f"error: invalid {name} str field: {e}")

    return s


def logon(host: str, port: int, timeout: float | None = None) -> str:
    # Send 'logon' command to cgminer and get the response
    timeout = TIMEOUT if timeout is None else timeout
    res = send_cgminer_simple_command(host, port, "logon", timeout)
    if res["STATUS"][0]["STATUS"] == "E":
        raise MinerCommandSessionAlreadyActive(host, port, res["STATUS"][0]["Msg"])
    # Check if the response has the expected structure
    session = validate_message(host, port, res, "SESSION")[0]
    return str(session["SessionID"])


def logoff(
    host: str, port: int, sid: str, timeout: float | None = None
) -> dict[str, Any]:
    timeout = TIMEOUT if timeout is None else timeout
    res = send_cgminer_command(host, port, "logoff", sid, timeout)
    return validate_message(host, port, res)


def add_session_id_parameter(session_id, parameters) -> list[Any]:
    # Add the session id to the parameters
    return [session_id, *parameters]


def parameters_to_string(parameters):
    # Convert the parameters to a string that LuxOS API accepts
    return ",".join(parameters)


def execute_command(
    host: str,
    port: int,
    timeout_sec: int | float | None,
    cmd: str,
    parameters: str | list[Any] | dict[str, Any] | None = None,
    verbose: bool = False,
):
    timeout_sec = TIMEOUT if timeout_sec is None else timeout_sec
    # Check if logon is required for the command
    logon_req = logon_required(cmd)

    parameters = parameters_to_list(parameters)

    if logon_req:
        # Get a SessionID
        sid = logon(host, port, timeout_sec)
        # Add the SessionID to the parameters list at the left.
        parameters = add_session_id_parameter(sid, parameters)

        log.debug("SessionID obtained for %s: %s", host, sid)

    # TODO verify this
    elif not logon_req:
        log.debug("Logon not required for executing %s", cmd)

    # convert the params to a string that LuxOS API accepts
    param_string = ",".join(parameters)

    log.debug("%s on %s with parameters: %s", cmd, host, param_string)

    # Execute the API command
    res = send_cgminer_command(host, port, cmd, param_string, timeout_sec)

    log.debug(res)

    # Log off to terminate the session
    if logon_req:
        logoff(host, port, sid, timeout_sec)

    return res


def rexec(
    host: str,
    port: int,
    cmd: str,
    parameters: str | list[Any] | dict[str, Any] | None = None,
    timeout: float | None = None,
    retry: int | None = None,
    retry_delay: float | None = None,
) -> dict[str, Any] | None:
    if retry or retry_delay:
        raise NotImplementedError("cannot use rexec in suncops with retry!")
    return execute_command(host, port, timeout, cmd, parameters)


@contextlib.contextmanager
def with_atm(host, port, enabled: bool, timeout: float | None = None):
    res = rexec(host, port, "atm", timeout=timeout)
    if not res:
        raise MinerConnectionError(host, port, "cannot check atm")
    current = validate_message(host, port, res, "ATM")[0]["Enabled"]
    rexec(host, port, "atmset", {"enabled": enabled}, timeout=timeout)
    yield current
    rexec(host, port, "atmset", {"enabled": current}, timeout=timeout)



"""Script to issue commands to miner(s)

This tool is designed to issue a simple command (with parameters) to a set of miners
defined on the command line (--range/--range_start/--range-end) or in a file (yaml/csv).
"""
from __future__ import annotations
import asyncio
import contextlib
import logging
import argparse
from pathlib import Path

from ..cli import v1 as cli
from ..syncops import execute_command

log = logging.getLogger(__name__)


def add_arguments(parser: cli.LuxosParserBase) -> None:

    def add_miners_arguments(group):
        group.add_argument("--range_start", help="IP start range")
        group.add_argument("--range_end", help="IP end range")
        group.add_argument("--ipfile", default=Path("ips.csv"), type=Path,
            help="File name to store IP addresses"
        )

        group.add_argument("--range", action="append", dest="addresses",
                           help="IPs range or @file", type=cli.flags.type_range)

        group.add_argument("--luxos-port", dest="luxos_port",
                           help="miners' default port", type=int, default=4028)
    group = parser.add_argument_group("Miners", "miners list or range")

    add_miners_arguments(group)
    cli.flags.add_arguments_rexec(parser)

    parser.add_argument("--cmd", "--command", dest="cmd", required=True, help="Command to execute on LuxOS API")
    parser.add_argument(
        "--params",
        dest="parameters",
        action="append",
        help="Parameters for LuxOS API (either str or key=value pair)",
    )
    parser.add_argument("--batch", dest="batchsize",
                        type=int, default=100,
                        help="limit execution to batch miners at time")

    group1 = parser.add_mutually_exclusive_group()
    group1.add_argument(
        "-a",
        "--all",
        dest="details",
        action="store_const",
        const="all",
        help="show full result output",
    )
    group1.add_argument(
        "-j",
        "--json",
        dest="details",
        action="store_const",
        const="json",
        help="show results in json format",
    )


def process_args(args: argparse.Namespace):

    def process_addresses():  # yes, it is pretty long
        from luxos.ips import iter_ip_ranges, load_ips_from_csv
        args.addresses = [
            address for addresses in (args.addresses or []) for address in addresses
        ]

        # old way
        if args.range_start and args.range_end:
            args.addresses.extend(iter_ip_ranges(f"{args.range_start}-{args.range_end}"))
        elif args.range_start:
            args.addresses.append((args.range_start, None))
        elif args.range_end:
            args.error("--range_end requires --range_start")

        if args.ipfile.exists():
            args.addresses.extend(load_ips_from_csv(args.ipfile))

        args.addresses = [
            (host, port or args.luxos_port)
            for host, port in args.addresses
        ]

    process_addresses()
    if not args.addresses:
        args.error("need one of the miners group flags (eg. --range_{start|end}/--range")

    lparameters : list[str] = []
    dparameters : dict[str, str] = {}

    for param in args.parameters or []:
        if "=" in param:
            key, _, value = param.partition("=")
            if lparameters:
                args.error("--params can be all strings or key=value pairs, not mixed")
            dparameters[key] = value
        else:
            if dparameters:
                args.error("--params can be all strings or key=value pairs, not mixed")
            lparameters.append(param)
    args.parameters = lparameters or dparameters


@cli.cli(add_arguments=add_arguments, process_args=process_args)
async def main(args: argparse.Namespace):
    from . import async_luxos
    await async_luxos.run(
        args.addresses,
        cmd=args.cmd, params=args.parameters,
        batchsize=args.batchsize, delay=2.0, details=args.details
    )


if __name__ == "__main__":
    asyncio.run(main())



import asyncio


class LuxosBaseException(Exception):
    pass


class MinerConnectionError(LuxosBaseException):
    def __init__(self, host: str, port: int, *args, **kwargs):
        super().__init__(host, port, *args, **kwargs)
        self.address = (host, port)

    def __str__(self):
        msg = "unknown reason"
        if getattr(self, "__cause__"):
            msg = repr(self.__cause__)
        elif self.args[2:]:
            msg = str(self.args[2])
        return (
            f"<{self.address[0]}:{self.address[1]}>: {self.__class__.__name__}, "
            f"{msg}"
        )


class MinerCommandTimeoutError(MinerConnectionError, asyncio.TimeoutError):
    pass


class MinerCommandSessionAlreadyActive(MinerConnectionError):
    pass


class MinerCommandMalformedMessageError(MinerConnectionError):
    pass


class LuxosLaunchError(MinerConnectionError):
    def __init__(self, tback: str, host: str, port: int, *args, **kwargs):
        self.tback = tback
        super().__init__(host, port, *args, **kwargs)

    def __str__(self):
        from .text import indent

        msg = indent(str(self.tback), "| ")
        return f"{self.address}: \n{msg}"


class LuxosLaunchTimeoutError(LuxosLaunchError, asyncio.TimeoutError):
    pass
